/*
  h_file.c   Nov98  JS  version 0.3
  
  this is a file for handling Htmp-File (tmp_hxy.000) 
  reasons and goals for this file:
  - 2GB file breaking => use LFS (64bit) or more blocks
  - error handling (perror nach printf???) Jun99
  - open/close
  - hidden variables (encapsulating of file-handling)
  - better reading of main-file
  Ideas:
  - chksum for tmp-file, check during read?
  - vvv==1 ? measure disk performance
  - compressed tmp_hxy.XXX ? (example: 1GB => 336MB, zip=17m  unzip=3m read=1m)
                        thor  gzip -c1 1GB => 640MB, zip=10m  unzip=3m
                        thor  gzip -c9 1GB => 579MB, zip=125m unzip=3m 
                        thor  gzip -c4 1GB => 550MB, zip=14m  unzip=3m
  - see pipe.c!!! (h_file,fopen,fclose,... & fn,pipe_ropen,... kapseln
  - repeat H*v if error!?
  - see also utils/vbuf.c for buffering/speed/IO
  
  this file is still changing Nov98,Apr99,Jun99,Nov99
  history:
   v2.17 - array of thxy -to- thxy of arrays
   v2.0.1 - access to mcplx only via Macros M_d2f(), M_f2d()
   v2.0.0 - file-split removed (see history)
   03.02.00 - bug cplx rr => mcplx rr
   25.11.99 - bug in my_fread/write if file > 1MB beseitigt
   07.01.00 - bug auf MIPS beseitigt (high-low Format!!!)
               Vorsicht bei type-Umwandlungen (high-low/low-high format)
   ToDo: remove pipe.c (simpler), add mem-only-version (mmap?) /dev/shm???

*/
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>	/* strerror */
#include "defs.h"
#include "spins.h"
#include "h_file.h"
#include "pipe.h"
#include "block.h"
#if 1
#include <unistd.h>		/* test this with configure! */
#define SLEEP(x) sleep(x)
#else
#define SLEEP(x) {}
#endif

/*
 * v2.30 can have different numbers of I/O threads and compute threads
 */

#ifndef HRMAX
#if HRMAX>0
#error "HRMAX!=0 if not defined?" 
#endif
#endif
#if HRMAX>0	/* table for 16bit index to Hamilton-entries */
 int  hr_len[B_NUM];    /* number of table entries */
 cplx hr[B_NUM][HRMAX];	/* table of matrix values, one table per thread */
 /* better use realloc for dyn. pt_n ? */
#endif

/* define the basename for the H-files htmpBBB.PPP
 * where BBB is the IO-thread-number (block)
 * and PPP the paramter index (ToDo)
 */
#define Htmp "./tmp/htmp"	/* ./tmp can be a link to /tmp/$USER */

/* public variables */

/* privat variables */
char  h_name[B_NUM][40]; /* file-name B_NUM=(ToDo: <=)B_NUM files */
int   h_error=0;         /* set if H could not be written */
int   h_state[B_NUM];    /* 0=closed, 1=ropen, 2=wopen */
int   h_open[B_NUM];    /* number cthreads have files opened 0..B_NUM/F_NUM-1 */
static int out_vvv=0;   /* number of debug outputs */


/* my comfortable erroroutput for misterious errors */
int my_ferror(char *msg){
  perror(msg);	/* errno_msg => stderr */
  SLEEP(1);
  return 0;
}


/* including error-handling (public???) */
size_t my_fread( char *ptr, size_t size, size_t nmemb, int block){
  size_t n_readed=0,nf,ni,tries=0; /* sgi-CC can not add void* */
  nmemb*=size;
  if (block>=B_NUM) ERR(9999);
  if( nmemb )
  while(n_readed< size /* nmemb */ ){ /* mind. 1 element */
    /* some clibs do alloc a buffer for disk_read which cost time if it is to big */
    ni=nmemb-n_readed;
    /* v2.29 add tries>0 because I forgot whats the splitting is good for */
    if(tries>0 && ni>262144) ni=262144;	/* 2^n=256kB pieces, why? */
    nf=pipe_read(ptr,(int)ni,block);
    if(nf<=0){ E3("read=%d ni=%d block=%d ",(int)nf,(int)ni,block);
               my_ferror("read");if(tries++<5)continue;break; }
    n_readed+=nf; ptr+=nf; tries=0;
  } /* while */
  return n_readed/size;
}

/* 
   write large objects in small 1e6-pieces,
    because fwrite (probably) need lot of memory
 */
size_t my_fwrite( void *ptr, size_t size, size_t nmemb, int block){
  size_t n_written=0, nf, ni, tries=0;
  char *p=(char *)ptr;	/* sgi-CC can not add void* */
  if (vvv&16 && out_vvv<1000) { out_vvv++;
    printf("\nvvv&16 fwrite %3d %d*%d %p err=%d ",
           mpi_i*pt_n+block, (int)size, (int)nmemb, ptr, h_error); }
  if (block>=B_NUM) ERR(9999); /* cant be! */
  if (h_state[block]!=2) ERR(9999); /* file not open */
  nmemb*=size;
  if (h_error) return 0;  /* dont try if something went wrong on I/O */
  while (n_written<nmemb) {
    ni=nmemb-n_written;
    /* v2.29 I forgot whats splitting good for, try to speedup */
    if(tries>0 && ni>262144) ni=262144;	/* 256kB pieces */
    nf=pipe_write(p,(int)ni,block);
    n_written+=nf; p+=nf;
    if(nf<ni){
      ERR(620);
      if (h_error) break;  /* dont try 5 times on old errors */
      if (tries++<5) continue; /* try 5 times */
      h_error=1; break;    /* write finally failed */
    }
    tries=0;
  } /* end while */
  /* if (vvv&16) printf(" %d/%d -",n_written,size); */
  return n_written/size;
}

/* --- end of file handler --- */

/* --------- interface between compute blocks and H storage ------ */

/* open for reading */
int h_ropen(int block){ /* 0..B_NUM-1, block > B_NUM possible */
  if (vvv&16) printf("\nvvv&16 h_ropen  %3d", mpi_i*pt_n+block);
  if (block<0 || block>=pt_n) ERR(9999); /* for safety */
  if (++h_open[block%B_NUM]==1) { /* open stream for first block */
    if (vvv&16) printf(" open %s", h_name[block%B_NUM] );
    if ( pipe_ropen(h_name[block],block) ){
      E1("fopen_%s",h_name[block]);return 1; }
  }
  b_hbl[block]=HBLen; /* force h_read on first h_get */
  if (h_state[block]!=0) { F1("h_ropen(%d): file was open", block); }
  h_state[block]=1;
  return 0;
}

/* open for writing */
int h_wopen(int block){  /* 0..B_NUM-1, block > B_NUM possible */
  if (block<0 || block>=pt_n) ERR(9999); /* for safety */
  if(vvv && mpi_i*pt_n+block==0)
    printf(" write " Htmp "000-%03d", mpi_n*pt_n-1);
  if (vvv&16) /* maximum info */
     printf("\nvvv&16 h_wopen  %3d open=%d", mpi_i*pt_n+block, h_open[block%B_NUM]);
  if (++h_open[block%B_NUM]==1) { /* open stream for first block */
    if (vvv&16) printf(" pipe_wopen %2d", block%B_NUM);
    if ( pipe_wopen(h_name[block%B_NUM],block%B_NUM) ) { 
      E1("wopen %s",h_name[block%B_NUM]); h_error=1; return 1; }
  }
  b_hbl[block]=0;
  if (h_state[block]!=0) ERR(9999); /* file not open */
  h_state[block]=2;
  return 0;
}

int h_rclose(int block) {
  if (vvv&16)
    printf("\nvvv&16 h_rclose %3d open=%d", mpi_i*pt_n+block, h_open[block%B_NUM]);
  if (h_state[block]!=1) ERR(9999); /* file not open */
  h_state[block]=0;
  if (--h_open[block%B_NUM]==0) /* close stream if last block flushed */
    pipe_close(block%B_NUM);
  return 0;
}

/* --- read from Htmp-file, ret: 0=ok --- 
 * read from disk into H-buffer, struct thxy = xy[HBLen], r[HBLen](or idx+para)
 *  ToDo: get special block from stream of random(?) stored blocks 
 *  search dynamic buffer, if not found extend buffer until block found
 *  max. B_NUM/B_NUM (* PARAM)?
 * this buffer is written by B_NUM IO-threads
 *   and read by B_NUM compute threads (v2.26 deadlocks?)
 * IO-thread=file blocks=compute-threads
 *                          v-blocks IO=0-Buffer? ToDo
 *    0:          0,1,4,0,5,0,1,4 (compute block 0 was faster than the others?)
 *    1:          2,3,6,7,2,3,6,7
 */
 
#if defined(STOREH) && STOREH==0
void *pmnext(int block); /* pipe.c */
#endif
/* privat */
/* read special block, !!! ToDo: read next block, caller must handle that
 *   read from stream[blk] or memory[blk] to tbh (i.e. b_hbuf[block])
 *   for STOREH==0 we simply return tbh as pointer to next H_block 
 */
struct thxy *h_read(struct thxy *tbh, int block){
#if defined(STOREH) && STOREH==0
  /* pipe_read((char *)tbh, Hsize, block); # old v2.35 */
  return ((struct thxy *)pmnext(block)); /* v2.36 Jul08 */
#else
  if (vvv&16 && out_vvv<1000)  /* print before, because read may hang */
    printf("\nvvv&16 h_read %3d state=%d", mpi_i*pt_n+block, h_state[block]);
  if (h_state[block]!=1) ERR(9999); /* file not open */
  if (my_fread((char *)&b_hbuf[block],Hsize,1,block%B_NUM)!=1){
     E0("h_read"); ERR(9622); return 1;
   }
  if (vvv&16 && out_vvv<1000) { out_vvv++;  /* move to h_read */
    printf("\nvvv&16 h_read %3d state=%d"
           " bj %2d j %6ld r %3d jf %d ...",
          mpi_i*pt_n+block, h_state[block],
           (int) b_hbuf[block].by[ 0 ],
          (long) b_hbuf[block].xy[ 0 ]&~XY_FLAG,
           (int) b_hbuf[block].rr[ 0 ],
          (int) ((b_hbuf[block].xy[ 0 ] & XY_FLAG)?1:0)); }
  return tbh;
#endif
}


/* ------------------------------------------------------------------ *
 * h_put: write matrix-element and its column (line encoded)          *
 * h_get: read  matrix-element and its column (line+1 if column>n1)   *
 * ------------------------------------------------------------------ */
#define HR_SIZE   sizeof(HR_TYPE)
#define XY_SIZE   sizeof(XY_TYPE)  


#if HRMAX>0
/* seperate error routine, only loaded to CPU-cache in error-case */
void err_fulltable(){
    int i, j;
    /* output maximum on information, what was went wrong? */
    printf("\nerr_fulltable: --debug--\n");
    for (i=0;i<pt_n;i++) if (hr_len[i]>=HRMAX-4) {
      printf(" Hblock=[%d.%d] b_hnz=%9ld hr_len=%4d hr[]=",
             mpi_i, i, b_hnz[i], hr_len[i]);
      for (j=0;j<hr_len[i] && j<100;j++) printf(" %g",D(hr[i][j]));
      if (j>=100) printf(" ...");
      printf("\n");
    }
    ERR(9100);
}
#endif

/* h_put fills a write buffer, this fkt writes the buffer out!
 * h_wclose is only called to close the file after writing
 * block=1..B_NUM-1 can be equal and greater than F_NUM to flush all buffers
 */
int h_wclose(int block){  /* store in file */
  if(h_error) return 1;
  if (vvv&16)
    printf("\nvvv&16 h_wclose %3d open=%d", mpi_i*pt_n+block, h_open[block%B_NUM]);
  if(b_hbl[block]){ /* buffer filled? */
   int i; /* fill rest of buffer, just to be sure and better debugging */
   for (i=b_hbl[block];i<HBLen;i++){
     b_hbuf[block].xy[i]=~0;
     b_hbuf[block].by[i]=~0;
     b_hbuf[block].rr[i]=0;
   }
   // printf(" hwrite %d (close)",block%B_NUM);
   if (h_state[block]!=2) ERR(9999); /* file not open */
   if (vvv&16)
     printf("\nvvv&16 h_write %3d", mpi_i*pt_n+block);
   if( my_fwrite((char *)&b_hbuf[block], Hsize, 1, block%B_NUM) != 1 ) {
     h_error=1; return 1;
   }
  }
  if (h_state[block]!=2) ERR(9999); /* file not open */
  b_hbl[block]=0;
  h_state[block]=0; /* mark as closed */
  if (--h_open[block%B_NUM]==0) { /* close stream if last block flushed */
    if (vvv&16) printf(" pipe_close %3d", (mpi_i*pt_n+block)%B_NUM);
    pipe_close(block%B_NUM); }
  return 0;
}

/* h_put and h_get moved to spins.c
 * to make the call shorter and allow inlining => no speedup Jan07 */

/* initialize H-storage, called by spins.c storeh*(), thread0 only */
void ini_thxy(ULong n1){ /* alloc */
  static int num_ini=0;   /* malloc only once */
  int i1,i2,j; ULLong i; XY_TYPE xy; struct thxy t1;
  if (!num_ini) {
     // if (!b_hbuf) b_hbuf=(struct thxy *)malloc(pt_n*sizeof(struct thxy));
     // if (!b_hbuf) ERR(9999);
  } num_ini++;
  for(j=0;j<pt_n;j++){
    h_state[j]=0; /* default 0=close */
#if HRMAX>0
    hr_len[j]=0; /* table for index-to-matrix value */
#endif
  }
  for(i1=0;i1<B_NUM;i1++){
    h_open[i1]=0;               /* no bocks open */
    h_name[i1][0]=0;		/* file-name ToDo: set name here! */
    sprintf(h_name[i1],Htmp "%03u.%03u",(int)i1+mpi_i*B_NUM,1);
    /* ToDo: set by user? spread over disks, better use links */
/*    for (i2=0;i2<pt_n;i2++)
      hblockbufs[i1][i2].block=-1;
*/ /* mark as free, ToDo */
  }
  /* measure: i1 = used bytes for xy + XY_FLAG,
   *          i2 = sizeof xy in bytes (ToDo: check sense)
   * ToDo: since v2.19 we store y-b_ofs[blk] (not y), less bits
   */
  for (i1=0,i=2*n1;i!=0;i1++,i>>=8); /* endless if i is signed <0 */
  for (i2=0,xy=1;xy!=0;i2++,xy<<=8);
  if (vvv && !mpi_i) 
    printf("\n ini_thxy: bytes(thxy.xy) %d bytes(n1) %d", i2, i1);
  if (  i1> i2) F0("xy to small, enlarge XY_TYPE in h_file.h");
  if (2*i1<=i2 && n1>100000 && !mpi_i) /* dont warn on play systems */
    printf("\nWarning: change XY_TYPE in h_file.h to spare memory.");
  j=sizeof(int); /* compiler warnings for if(const1==const2) */
  if (j!=sizeof(t1.xy[0])) E0("sizeof int!=thxz.xy");
}

#define HR_FILE "tmp/hr_save%03d.dat"  /* ToDo: mpi .001 etc. */

/* checkpoint Hstorage, write restore data to HR_FILE ToDo: mpi */
void h_checkpoint(int vvv){
  char s1[80];
  int i, hrmax;
  FILE *f1=NULL;
  /* if( save_mode & 1 ) --- small amount of data! save every time */
  { /* save data before possible crash */
    sprintf(s1,HR_FILE,mpi_i);
    f1=fopen(s1,"wb"); if(!f1)E1("error fopen %s",s1); }
  for(hrmax=i=0;i<pt_n;i++){
#if HRMAX>0
    if (hr_len[i]>hrmax) hrmax=hr_len[i];
    if (f1){
      /* ToDo: better outpiut num threads*mpi + hr_min/max */
      if (vvv)
        printf(((i==0 && !mpi_i)?" save hr[] len: %d":" %d"),hr_len[i]);
      fwrite(&hr_len[i],sizeof(hr_len[i]), 1, f1);
      fwrite(&hr[i][0], sizeof(hr[i][0]), hr_len[i], f1); /* hr_len==0 ??? */
    }
#endif
    fwrite(b_hnz, sizeof(b_hnz[0]), pt_n, f1);
    fwrite(b_snz, sizeof(b_snz[0]), pt_n, f1);
    fwrite(b_sn1, sizeof(b_sn1[0]), pt_n, f1);
  }
#ifdef CONFIG_MPI
  MPI_Allreduce(&hrmax,&i,1,MPI_INT,MPI_MAX,MPI_COMM_WORLD); hrmax=i;
#endif
  if (f1) fclose(f1);
  if (vvv && !mpi_i) {
    printf("\n h_checkpoint: max hr_num= %5d (entries/block) %s..\n",
           hrmax, s1);
  }
}

/* restore H-storage, after h_checkpoint */
void h_restore(int vvv){
  char s1[80];
  int i, hrmax;
  FILE *f1=NULL;
  /* load data after crash */
  sprintf(s1,HR_FILE,mpi_i);
  f1=fopen(s1,"rb");if(!f1)E0("error fopen");
#if HRMAX>0
  for(hrmax=i=0;i<pt_n;i++){
    if(vvv)printf(" load hr[%d]=",i);
    fread(&hr_len[i],sizeof(hr_len[i]),1,f1);
    fread(&hr[i][0], sizeof(hr[i][0]), hr_len[i],f1);
    if( hr_len[i]>hrmax ) hrmax=hr_len[i];
    if(vvv)printf("%d",hr_len[i]);
  }
#endif
  fread(b_hnz, sizeof(b_hnz[0]), pt_n, f1);
  fread(b_snz, sizeof(b_snz[0]), pt_n, f1);
  fread(b_sn1, sizeof(b_sn1[0]), pt_n, f1);
  if(f1)fclose(f1);
#ifdef CONFIG_MPI
  MPI_Allreduce(&hrmax,&i,1,MPI_INT,MPI_MAX,MPI_COMM_WORLD); hrmax=i;
#endif
  if(vvv){ printf(" h_restore: max hr_num= %5d (entries/block)\n",hrmax); }
}
/* ---------------------------------------------------------- */
